import base64
import json

from Crypto.Hash import SHA256, SHA512
from Crypto.Signature import PKCS1_v1_5
from datetime import datetime, timezone
from typing import Tuple, Union

from .enums import AlgorithmType, KeyType, SignatureAlgorithm
from .errors import SignatureFailureError


HASHES = {
	'sha256': SHA256,
	'sha512': SHA512
}

ALGORITHMS = {
	'hs2019',
	'rsa-sha1',
	'rsa-sha256',
	'ecdsa-sha256'
}


def generate_body_digest(body, algorithm='sha256'):
	if isinstance(body, str):
		body = body.encode('utf-8')

	h = HASHES[algorithm].new(data=body)
	return base64.b64encode(h.digest()).decode('utf-8')


def create_signature_algorithm(sigalg, hashalg):
	sigalg = KeyType.parse(sigalg)
	hashalg = parse_hashalg(hashalg)

	if hashalg == 'hs2019':
		return hashalg

	return f'{sigalg.value}-{hashalg}'


def parse_hashalg(hashalg):
	return hashalg.lower().replace('-', '')


class Digest:
	'Represents a body digest'

	__slots__ = {'digest', 'algorithm'}

	def __init__(self, digest, algorithm):
		'''
			Create a new digest object from an already existing digest

			:param str digest: Base64 encoded hash
			:param str algorithm: Algorithm used for hash in the format of ``[type]-[bytes]``
		'''

		self.digest = digest
		self.algorithm = algorithm


	def __repr__(self):
		return f'Digest({self.digest}, {self.algorithm})'


	def __str__(self):
		return self.compile()


	@classmethod
	def new(cls, body, size:int=256) -> 'aputils.Digest':
		'''
			Generate a new body digest

			:param typing.Union[dict,bytes,str] body: Message body to hash
			:param typing.Optional[int] size: SHA hash size
		'''

		if isinstance(body, dict):
			body = json.dumps(body).encode('utf-8')

		elif isinstance(body, str):
			body = body.encode('utf-8')

		if size >= 1024:
			size = int(size / 8)

		digest = generate_body_digest(body, f'sha{size}')
		return cls(digest, f'SHA-{size}')


	@classmethod
	def new_from_digest(cls, digest) -> 'aputils.Digest':
		'''
			Create a new digest from a digest header

			:param str digest: Digest header
		'''
		if not digest:
			return

		alg, digest = digest.split('=', 1)
		return cls(digest, alg)


	@property
	def hashalg(self):
		return self.algorithm.replace('-', '').lower()


	def compile(self) -> str:
		'Turn the digest object into a ``str`` for the Digest header'
		return '='.join([self.algorithm, self.digest])


	def validate(self, body) -> bool:
		'''
			Check if the body digest matches the body

			:param typing.Union[dict,bytes,str] body: Message body to verify
		'''
		if isinstance(body, dict):
			body = json.dumps(body).encode('utf-8')

		elif isinstance(body, str):
			body = body.encode('utf-8')

		digest = generate_body_digest(body, self.hashalg)

		return digest == self.digest


class Dotdict(dict):
	'A dict but items can be accessed as properties'

	def __init__(self, data:Union[dict,list,tuple]=None, **kwargs):
		dict.__init__(self)

		if isinstance(data, (list,tuple,set)):
			data = dict(data)

		self.update(data, **kwargs)


	def __setitem__(self, key, value):
		if isinstance(value, dict) and not isinstance(value, Dotdict):
			value = Dotdict(value)

		elif key == 'links' and isinstance(value, (list,tuple,set)):
			value = [Dotdict(v) for v in value]

		dict.__setitem__(self, key, value)


	def __getattr__(self, key):
		if key in self:
			return self[key]

		return object.__getattribute__(self, key)


	def __setattr__(self, key, value):
		if key in self or not key.startswith('_'):
			return self.__setitem__(key, value)

		object.__setattr__(self, key, value)


	def __setattr__(self, key):
		if key in self or not key.startswith('_'):
			return self.__delitem__(key)

		object.__delattr__(self, key)


	@classmethod
	def new_from_json(cls, data) -> 'aputils.Dotdict':
		'''
			Parse a JSON string to a dict

			:param Union[str,bytes] data: JSON string to parse
		'''
		return cls(json.loads(data))


	def to_json(self, indent=None) -> str:
		'''
			Turn the dict into a JSON string

			:param typing.Optional[typing.Union[str,int]] indent: Number of spaces (:class:`int`) or character (:class:`str`) to use for indentation
		'''
		return json.dumps(self, indent=indent)


	## re-implement dict.update so __setitem__ gets called
	def update(self, data:Union[dict,list,tuple]=None, **kwargs):
		if not data:
			data = {}

		for key, value in data.items():
			self[key] = value

		for key, value in kwargs.items():
			self[key] = value


class HttpDate(datetime):
	'Datetime object with convenience methods to parsing and creating HTTP date strings'

	FORMAT:str = '%a, %d %b %Y %H:%M:%S GMT'
	'Format to pass to datetime when (de)serializing a raw HTTP date'

	def __str__(self):
		return self.to_string()


	@classmethod
	def parse(cls, date) -> 'aputils.HttpDate':
		'''
			Parse an HTTP date in string format

			:param str date: String from an HTTP Date header
		'''
		data = cls.strptime(date, cls.FORMAT)
		return data.replace(tzinfo=timezone.utc)


	@classmethod
	def new_from_datetime(cls, date):
		data = cls.fromisoformat(date.isoformat())
		return data.replace(tzinfo=timezone.utc)


	@classmethod
	def new_utc(cls):
		data = cls.utcnow()
		return data.replace(tzinfo=timezone.utc)


	def timestamp(self) -> int:
		'Return the date as a unix timestamp without microseconds'
		return int(datetime.timestamp(self))


	def to_string(self) -> str:
		'Create an HTTP Date header string from the datetime object'
		return self.strftime(self.FORMAT)


class Signature:
	'Represents a signature header with access to values as attributes'

	__slots__ = {'keyid', 'algorithm', 'headers', 'signature', 'created', 'expires'}

	def __init__(self, signature, keyid, algorithm, headers, created=None, expires=None):
		'''
			Create a new signature object. This should not be initiated directly.

			:param str signature: Generated signature hash
			:param str keyid: URL of the public key
			:param str algorithm: Hashing and signing algorithms used to create the signature
			:param typing.Union[list,str] headers: Header keys used to create the signature
			:param typing.Optional[int] created: Unix timestamp representing the signature creation date
			:param typing.Optional[int] expires: Unix timestamp representing the date the signature expires
		'''
		self.signature:str = signature
		'Generated signature hash'
		self.keyid:str = keyid
		'URL of the public key'
		self.algorithm:SignatureAlgorithm = algorithm
		'Hashing and signing algorithms used to create the signature'
		self.headers:list = headers
		'Header keys used to create the signature'
		self.created:int = created
		'Unix timestamp representing the signature creation date'
		self.expires:int = expires
		'Unix timestamp representing the date the signature expires'


	def __setattr__(self, key, value):
		if key == 'headers':
			if isinstance(value, str):
				value = value.split()

		elif key == 'algorithm':
			value = SignatureAlgorithm.parse(value)

		elif key in {'created', 'expires'} and value != None:
			value = int(value)

		object.__setattr__(self, key, value)


	@classmethod
	def new_from_headers(cls, headers:dict[str,str]) -> 'aputils.Signature':
		'''
			Parse the signature from a header dict

			:param dict[str,str] headers: Header key/value pairs
			:raises KeyError: When the signature header(s) cannot be found
		'''

		headers = {key.lower(): value for key, value in headers.items()}

		if 'signature-input' in headers:
			raise NotImplementedError('Newer signature spec not supported yet')

		signature = headers['signature']
		data = {}

		for chunk in signature.strip().split(','):
			key, value = chunk.split('=', 1)
			data[key.lower()] = value.strip('\"')

		return cls(**data)


	@classmethod
	def new_from_signature(cls, string:str) -> 'aputils.Signature':
		'''
			Parse a Signature header

			DeprecationWarning: use :meth:`aputils.Signature.new_from_headers` instead

			:param str string: Signature header string
		'''

		return cls.new_from_headers({'signature': string})


	@property
	def algs(self) -> Tuple[str,str]:
		return self.algorithm.split('-', 1)


	@property
	def hashalg(self) -> str:
		'Algorithm used for hashing'
		return self.algs[1]


	@property
	def signalg(self) -> str:
		'Algorithm used for signing'
		return self.algs[2]


	@property
	def algorithm_type(self) -> AlgorithmType:
		'Type of algorithm used for this signature'
		if self.algorithm.lower() == 'hs2019':
			return AlgorithmType.HS2019

		elif self.algorithm.lower().startswith(('ecdsa', 'rsa')):
			return AlgorithmType.ORIGINAL


	def compile(self) -> str:
		'Generate a string for a Signature header'

		data = {
			'keyId': self.keyid,
			'algorithm': self.algorithm.value,
			'headers': ' '.join(self.headers),
			'created': self.created,
			'expires': self.expires,
			'signature': self.signature
		}

		for key, value in [*data.items()]:
			if value == None:
				del data[key]

		return ','.join(['{}="{}"'.format(k, v) for k, v in data.items()])
