import base64
import builtins
import json
import warnings

from Crypto.PublicKey import RSA, ECC
from Crypto.Signature import PKCS1_v1_5, DSS
from datetime import datetime, timedelta
from functools import wraps
from pathlib import Path
from typing import Union, Any
from urllib.parse import urlparse
from urllib.request import Request

from .enums import AlgorithmType, KeyType
from .errors import SignatureFailureError
from .misc import HASHES, Digest, HttpDate, Signature, parse_hashalg

## todo: import various request classes for signing/verifying


## check if the key is private or not
def check_private(private):
	name = 'private' if private else 'public'

	def outer(func):
		@wraps(func)
		def wrapper(key, *args, **kwargs):
			if key.is_private != private:
				raise TypeError(f'Cannot use method "{func.__name__}" on Signer with {name} key')

			return func(key, *args, **kwargs)
		return wrapper
	return outer


## check which kind of key is being used
def check_key_type(*types):
	types = [KeyType.parse(type) for type in types]

	def outer(func):
		@wraps(func)
		def wrapper(key, *args, **kwargs):
			if key.algorithm not in types:
				raise TypeError(f'Cannot use method "{func.__name__}" on Signer with {key.algorithm.value} key')

			return func(key, *args, **kwargs)
		return wrapper
	return outer


def create_sigstring(headers, used_headers):
	string = '\n'.join(f'{key}: {headers[key]}' for key in used_headers)
	return string.encode('ascii')


class Signer:
	'Used to sign or verify HTTP headers'

	__test_mode = False


	def __init__(self, key, keyid):
		'''
			Create a new signer object. The key can be an ``RsaKey`` object, ``str``, or ``Path`` to an exported key

			:param typing.Union[Crypto.PublicKey.RSA.RsaKey,Crypto.PublicKey.ECC.EccKey,pathlib.Path,str] key: RSA or ECC key to use for signing or verifying
			:param str keyid: Url to a web resource which hosts the public key
		'''

		self.key:Union[ECC.EccKey,RSA.RSAKey] = key
		'Key to use for signing or verifying'
		self.keyid:str = keyid
		'Url to a web resource which hosts the public key'


	def __repr__(self):
		return f'{self.__class__.__name__}({self.size.name}, keyid="{self.keyid}")'


	def __setattr__(self, key, value):
		if key == 'key':
			if isinstance(value, Path):
				with value.open('r') as fd:
					value = fd.read()

			if isinstance(value, str):
				if not value.startswith('-'):
					with Path(value).expanduser().resolve().open('r') as fd:
						value = fd.read()

				try:
					value = RSA.import_key(value)

				except ValueError:
					value = ECC.import_key(value)

					if value.curve != 'Ed25519':
						raise TypeError('Invalid ECC key. Only the Ed25519 curve is supported.') from None

			elif not isinstance(value, (ECC.EccKey, RSA.RsaKey)):
				raise TypeError('Key must be an RsaKey, EccKey, Path, or a string representation of a key')

		object.__setattr__(self, key, value)


	# the current spec just uses sha256 for now, but keeping this function for future use
	def _hash_func(self, algorithm):
		return HASHES['sha256']
		#algorithm = AlgorithmType.parse(algorithm)
		#raise NotImplementedError(algorithm)


	def _sign_func(self):
		if self.type == KeyType.ECC:
			return DSS

		elif self.type == KeyType.RSA:
			return PKCS1_v1_5

		raise NotImplementedError(self.type)


	def _generate_signature(self, headers, used_headers, algorithm):
		sighash = self._hash_func(algorithm).new(data=create_sigstring(headers, used_headers))
		sigdata = self._sign_func().new(self.key).sign(sighash)

		return base64.b64encode(sigdata).decode('utf-8')


	def _validate_signature(self, headers, signature):
		return self._sign_func().new(self.key).verify(
			self._hash_func(signature.algorithm_type).new(data=create_sigstring(headers, signature.headers)),
			base64.b64decode(signature.signature)
		)


	@classmethod
	def new(cls, keyid, type=KeyType.RSA, size=4096) -> 'aputils.Signer':
		'''
			Create a new signer with a generated ``RsaKey`` of the specified size

			:param str keyid: Url to a web resource which hosts the public key
			:param typing.Union[str,KeyType] type: Type of private key to generate
			:param int size: Size of RSA key in bits to generate. This is ignore for ECC keys.
		'''
		type = KeyType.parse(type)

		if type == KeyType.RSA:
			return cls(RSA.generate(size), keyid)

		elif type == KeyType.ECC:
			return cls(ECC.generate(curve='Ed25519'), keyid)


	@classmethod
	def new_from_actor(cls, actor) -> 'aputils.Signer':
		'''
			Create a signer object from an ActivityPub actor dict

			:param dict actor: ActivityPub Actor object
		'''

		## todo: use Message
		return cls(actor['publicKey']['publicKeyPem'], actor['publicKey']['id'])


	@property
	@check_key_type(KeyType.RSA)
	def bits(self) -> int:
		'Size of the RSA key in bits'
		return self.key.size_in_bits()


	@property
	@check_key_type(KeyType.RSA)
	def bytes(self) -> int:
		'Size of the RSA key in bytes'
		return self.key.size_in_bytes()


	@property
	def is_private(self) -> bool:
		'Return ``True`` if the key is private'
		return self.key.has_private()


	@property
	def type(self) -> 'aputils.KeyType':
		'Algorithm used to generate the key'
		if type(self.key) == ECC.EccKey:
			return KeyType.ECC

		elif type(self.key) == RSA.RsaKey:
			return KeyType.RSA


	@property
	@check_private(True)
	def pubkey(self) -> str:
		'Export the public key to a str'
		key = self.key.public_key().export_key(format='PEM')

		try: return key.decode('utf-8')
		except: return key


	def export(self, path=None) -> str:
		'''
			Export the key to a str

			:param typing.Optional[typing.Union[Path,str]] path: Path to dump the key in text form to if specified
		'''
		key = self.key.export_key(format='PEM')

		try: key = key.decode('utf-8')
		except: pass

		if path:
			path = Path(path)

			with path.open('w') as fd:
				fd.write(key)

		return key


	@check_private(True)
	def sign_headers(self, method, url, data=None, headers=None, sign_all=False, algorithm=AlgorithmType.HS2019) -> dict[str,str]:
		'''
			Generate a signature and return the headers with a "Signature" key

			Note: HS2019 is the only supported algorithm, so only use others when you absolutely have to

			:param str method: HTTP method of the request
			:param str url: URL of the request
			:param typing.Optional[dict[str,typing.Any]] data: ActivityPub message for a POST request
			:param typing.Optional[dict[str,str]] headers: Request headers
			:param typing.Optional[bool] sign_all: If ``True``, sign all headers instead of just the required ones
			:param aputils.AlgorithmType algorithm: Type of algorithm to use for hashing the headers. HS2019 is the only non-deprecated algorithm.
		'''

		headers = {} if not headers else {key.lower(): value for key, value in headers.items()}
		date = headers.get('date', HttpDate.new_utc())

		if isinstance(date, str):
			date = HttpDate.parse(date)

		elif isinstance(date, datetime):
			if date.__class__ == datetime:
				date = HttpDate.new_from_datetime(datetime)

			headers['date'] = date.to_string()

		url = urlparse(url)
		used_headers = ['(request-target)', 'host', 'date']
		headers.update({
			'(request-target)': f'{method.lower()} {url.path}',
			'host': url.netloc
		})

		if algorithm == AlgorithmType.HS2019:
			signature_alg = 'hs2019'
			used_headers.extend(['(created)', '(expires)'])
			headers.update({
				'(created)': str(date.timestamp()),
				'(expires)': str((date + timedelta(hours=6)).timestamp())
			})

		else:
			signature_alg = f'{self.type.value}-sha256'

		if data:
			used_headers.extend(['digest', 'content-length'])
			data = json.dumps(data)
			headers.update({
				'digest': Digest.new(data).compile(),
				'content-length': str(len(data.encode('utf-8')))
			})

		if sign_all:
			for key in headers.keys():
				if key not in used_headers:
					used_headers.append(key)

		signature = Signature(
			keyid = self.keyid,
			algorithm = signature_alg,
			headers = used_headers,
			signature = self._generate_signature(headers, used_headers, algorithm)
		)

		if algorithm == AlgorithmType.HS2019:
			signature.created = headers['(created)']
			signature.expires = headers['(expires)']

		headers['signature'] = signature.compile()

		for key in [*headers.keys()]:
			if key.startswith('(') or key == 'host':
				del headers[key]

		return headers


	@check_private(True)
	def sign_request(self, request, algorithm=AlgorithmType.HS2019) -> Any:
		'''
			Convenience function to sign a request. Support for more Request classes planned.

			:param urllib.request.Request request: Request object to sign
			:param aputils.AlgorithmType algorithm: Type of algorithm to use for signing and hashing the headers. HS2019 is the only non-deprecated algorithm.

			:raises TypeError: If the Request class is not supported
		'''

		# sign built-in request class
		if isinstance(request, Request):
			request_headers = dict(request.header_items())
			headers = self.sign_headers(
				request.get_method().upper(),
				request.full_url, request.data,
				request_headers,
				algorithm = algorithm
			)

			request.headers = {key.title(): value for key, value in headers.items()}

		else:
			raise TypeError(f'Request from module not supported: {type(request).__module__}')

		return request


	@check_private(False)
	def validate_signature(self, method, path, headers, body=None) -> bool:
		'''
			Check to make sure the Signature and Digest headers match

			:param str method: Request method
			:param str path: Request path
			:param dict[str,str] headers: Request headers
			:param typing.Optional[typing.Union[bytes,str]] body: Request body if it exists

			:raises aputils.SignatureFailureError: When any step of the validation process fails
		'''

		headers = {key.lower(): value for key, value in headers.items()}
		headers['(request-target)'] = ' '.join([method.lower(), path])
		digest = Digest.new_from_digest(headers.get('digest'))

		try:
			sig = Signature.new_from_headers(headers)

		except KeyError:
			raise SignatureFailureError('Missing signature')

		if digest:
			if not body:
				raise SignatureFailureError('Missing body for digest verification')

			if not digest.validate(body):
				raise SignatureFailureError('Body digest does not match')

		if sig.algorithm_type == 'hs2019':
			## hs2019 requires (created)
			if '(created)' not in sig.headers:
				raise SignatureFailureError('"(created)" header not used')

			if not self._test_mode:
				current_timestamp = HttpDate.new_utc().timestamp()

				if sig.created > current_timestamp:
					raise SignatureFailureError('Creation date after current date')

				if current_timestamp > sig.expires:
					raise SignatureFailureError('Expiration date before current date')

			headers['(created)'] = sig.created
			headers['(expires)'] = sig.expires

		valid = self._validate_signature(headers, sig)

		if not valid:
			raise SignatureFailureError('Signature does not match')

		return valid


	@check_private(False)
	async def validate_aiohttp_request(self, request) -> bool:
		'''
			Validate the signature header of an AIOHTTP server request object
		'''
		return self.validate_signature(request.method, request.path, request.headers, await request.read())
